/*
 * Copyright (C) 1997-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file client.c the client program used to connect to para_server */

#include <regex.h>
#include <openssl/rc4.h>

#include "para.h"
#include "list.h"
#include "sched.h"
#include "crypt.h"
#include "client.cmdline.h"
#include "string.h"
#include "stdin.h"
#include "stdout.h"
#include "client.h"
#include "error.h"

INIT_CLIENT_ERRLISTS;

static struct client_task *ct;
static struct stdin_task sit;
static struct stdout_task sot;

static void supervisor_post_select(__a_unused struct sched *s, struct task *t)
{
	if (ct->task.error < 0) {
		t->error = ct->task.error;
		return;
	}
	if (ct->status == CL_SENDING) {
		stdin_set_defaults(&sit);
		sit.buf = para_malloc(sit.bufsize),
		register_task(&sit.task);
		ct->inbuf = sit.buf;
		ct->in_loaded = &sit.loaded;
		ct->in_error = &sit.task.error;
		t->error = -E_TASK_STARTED;
		return;
	}
	if (ct->status == CL_RECEIVING) {
		stdout_set_defaults(&sot);
		sot.bufp = &ct->buf;
		sot.loaded = &ct->loaded;
		sot.input_error = &ct->task.error;
		register_task(&sot.task);
		t->error = -E_TASK_STARTED;
		return;
	}
}

static struct task svt = {
	.post_select = supervisor_post_select,
	.status = "supervisor task"
};

static int client_loglevel; /* loglevel */
INIT_STDERR_LOGGING(client_loglevel);


/**
 * The client program to connect to para_server.
 *
 * \param argc Usual argument count.
 * \param argv Usual argument vector.
 *
 * It registers two tasks: The client task that communicates with para_server
 * and the standard out task that writes any output produced by the client task
 * to standard out.
 *
 * \return EXIT_SUCCESS or EXIT_FAILURE
 *
 * \sa client_open(), stdout.c, stdout.h, para_client(1), para_server(1)
 */
int main(int argc, char *argv[])
{

	int ret;
	static struct sched s;

	init_random_seed_or_die();
	s.default_timeout.tv_sec = 1;
	s.default_timeout.tv_usec = 0;
	ret = client_open(argc, argv, &ct, &client_loglevel);
	if (ret < 0) /* can not use PARA_LOG here because ct is NULL */
		exit(EXIT_FAILURE);
	register_task(&svt);
	ret = schedule(&s);
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	client_close(ct);
	return ret >= 0? EXIT_SUCCESS: EXIT_FAILURE;
}
