/*
 * Copyright (C) 2006-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/**
 * \file grab_client.c Functions for grabbing the stream at any position
 * in a filter chain.
 *
 * \sa filter_chain filter_chain_info filter.
 */

#include <regex.h>
#include <sys/types.h>
#include <dirent.h>

#include "para.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "filter.h"
#include "grab_client.h"
#include "audiod.h"
#include "error.h"
#include "string.h"
#include "fd.h"

/** Grab clients that are not yet attached to a filter node. */
static struct list_head inactive_grab_client_list;

static int max_num_filters(void)
{
	int i, ret = 0;

	for (i = 0; audio_formats[i]; i++) {
		PARA_INFO_LOG("%s filter chain length: %d\n", audio_formats[i],
			num_filters(i));
		ret = PARA_MAX(ret, num_filters(i));
	}
	PARA_INFO_LOG("maximal filter chain length: %d\n", ret);
	return ret;
}

static struct filter_node *find_filter_node(int format, int filternum)
{
	int i;

	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		if (s->format < 0 || !s->fc)
			continue;
		if (format >= 0 && s->format != format)
			continue;
		if (num_filters(i) <= filternum)
			continue;
		/* success */
		return  s->fc->filter_nodes + filternum;
	}
	return NULL;
}

static int gc_write(char *buf, size_t len, struct filter_callback *fcb)
{
	struct grab_client *gc = container_of(fcb, struct grab_client, fcb);
	size_t written = 0;

	while (written < len) {
		int ret = write_ok(gc->fd);
		if (ret < 0)
			goto err;
		if (ret == 0) { /* fd not ready */
			if (gc->mode == GM_PEDANTIC)
				goto err;
			if (gc->mode == GM_SLOPPY)
				return 1;
		}
		ret = write(gc->fd, buf + written, len - written);
		if (ret < 0) {
			if (errno != EAGAIN && errno != EINTR)
				goto err;
			if (gc->mode == GM_PEDANTIC)
				goto err;
			if (gc->mode == GM_SLOPPY)
				return 1;
		} else
			written += ret;
	}
	return 1;
err:
	gc->error = -E_GC_WRITE;
	return -E_GC_WRITE;
}

static void add_inactive_gc(struct grab_client *gc)
{
	PARA_INFO_LOG("adding grab client %p (fd %d) to inactive list\n",
		gc, gc->fd);
	para_list_add(&gc->node, &inactive_grab_client_list);
}

static void gc_close(struct filter_callback *fcb)
{
	struct grab_client *gc = container_of(fcb, struct grab_client, fcb);

	if ((gc->flags & GF_ONE_SHOT) || gc->error < 0) {
		PARA_INFO_LOG("closing fd %d (grab client %p)\n", gc->fd, gc);
		close(gc->fd);
		free(gc);
		return;
	}
	add_inactive_gc(gc);
}

/**
 * Move a grab client from the inactive list to a filter node.
 *
 * \param gc The grab client to activate.
 * \param fn The filter node \a gc gets attached to.
 *
 * \sa filter_node::callbacks, inactive_grab_client_list.
 */
void activate_grab_client(struct grab_client *gc, struct filter_node *fn)
{
	PARA_INFO_LOG("activating %p (fd %d, filter node: %p)\n", gc, gc->fd, fn);
	list_del(&gc->node);
	para_list_add(&gc->fcb.node, &fn->callbacks);
}

/**
 * Activate inactive grab clients if possible.
 *
 * \param audio_format_num The number of the audio format of the new audio file.
 * \param fc The filter chain containing the activated filters.
 *
 * This is called from audiod.c when the current audio file changes. It loops
 * over all inactive grab clients and checks each grab client's configuration
 * to determine if the client in question wishes to grab the new stream.  If
 * yes, this grab client is moved from the inactive grab client list to an
 * appropriate filter_node.
 *
 * \sa filter_chain_info::filters, inactive_grab_client_list,
 * activate_grab_client.
 */
void activate_inactive_grab_clients(int audio_format_num,
		struct filter_chain *fc)
{
	struct grab_client *gc, *tmp;
	struct filter_node *fn;

	list_for_each_entry_safe(gc, tmp, &inactive_grab_client_list, node) {
		if (gc->audio_format_num >= 0 && gc->audio_format_num !=
				audio_format_num)
			continue;
		if (gc->filter_num >= num_filters(audio_format_num))
			continue;
		fn = fc->filter_nodes + gc->filter_num;
		activate_grab_client(gc, fn);
	}
}

static int check_gc_args(int argc, char **argv, struct grab_client *gc)
{
	int i, ret;

	gc->audio_format_num = -1; /* default: grab any audio format */
	for (i = 1; i < argc; i++) {
		const char *arg = argv[i];
		if (arg[0] != '-')
			break;
		if (!strcmp(arg, "--")) {
			i++;
			break;
		}
		if (!strncmp(arg, "-n=", 3)) {
			ret = para_atoi32(arg + 3, &gc->filter_num);
			if (ret < 0)
				return ret;
			if (gc->filter_num < 0)
				return -E_BAD_GC_FILTER_NUM;
			if (gc->filter_num >= max_num_filters())
				return -E_BAD_GC_FILTER_NUM;
			continue;
		}
		if (!strncmp(arg, "-m", 2)) {
			if (*(arg + 3))
				return -E_GC_SYNTAX;
			switch(*(arg + 2)) {
			case 's':
				gc->mode = GM_SLOPPY;
				continue;
			case 'a':
				gc->mode = GM_AGGRESSIVE;
				continue;
			case 'p':
				gc->mode = GM_PEDANTIC;
				continue;
			default:
				return -E_GC_SYNTAX;
			}
		}
		if (!strcmp(arg, "-i")) {
			gc->flags |= GF_INPUT_GRAB;
			continue;
		}
		if (!strcmp(arg, "-o")) {
			gc->flags |= GF_ONE_SHOT;
			continue;
		}
		if (!strncmp(arg, "-f=", 3)) {
			ret = get_audio_format_num(arg + 3);
			if (ret < 0)
				return ret;
			gc->audio_format_num = ret;
			continue;
		}
		return -E_GC_SYNTAX;
	}
	if (i != argc)
		return -E_GC_SYNTAX;
	return 1;
}

/**
 * Check the command line options and allocate a grab_client structure.
 *
 * \param fd The file descriptor of the client.
 * \param argc Argument count.
 * \param argv Argument vector.
 *
 * If the command line options given by \a argc and \a argv are valid.
 * allocate a struct grab_client and initialize it with this valid
 * configuration. Moreover, add the new grab client to the inactive list.
 *
 * \return Standard.
 *
 * \sa grab_client, inactive_grab_client_list, activate_grab_client,
 * filter_node::callbacks.
 */
int grab_client_new(int fd, int argc, char **argv)
{
	int ret;
	struct grab_client *gc = para_calloc(sizeof(struct grab_client));
	struct filter_node *fn;

	ret = check_gc_args(argc, argv, gc);
	if (ret < 0)
		goto err_out;
	if (gc->flags & GF_INPUT_GRAB)
		gc->fcb.input_cb = gc_write;
	else
		gc->fcb.output_cb = gc_write;
	gc->fd = fd;
	gc->fcb.close = gc_close;
	fn = find_filter_node(gc->audio_format_num, gc->filter_num);
	if (fn)
		para_list_add(&gc->fcb.node, &fn->callbacks);
	else
		add_inactive_gc(gc);
	return 1;
err_out:
	free(gc);
	return ret;
}

/**
 * Initialize the grabbing subsystem.
 *
 * This has to be called once during startup before any other function from
 * grab_client.c may be used. It initializes \a inactive_grab_client_list.
 */
void init_grabbing(void)
{
	PARA_INFO_LOG("grab init\n");
	INIT_LIST_HEAD(&inactive_grab_client_list);
}
